"""
Exchange stream cache for market and order data.

This cache implementation tries to be efficient and minimalistic by using the following principles:
- Use builtin data structures whenever possible
- Minimize creation and allocation of new objects
- Avoid any kind of sorting

Though, there is probably quite some room for further optimizations.
"""

from collections import defaultdict

from betfair_parser.spec.streaming import (
    LPV,
    MCM,
    OCM,
    PV,
    ChangeType,
    MarketDefinition,
    MatchedOrder,
    Order,
    OrderRunnerChange,
    RunnerChange,
    SegmentType,
)


def ladder_update_lpv(ladder: dict[int, LPV], lpvs: list[LPV]) -> None:
    for lpv in lpvs:
        if not lpv.volume:
            ladder.pop(lpv.level, None)
        else:
            ladder[lpv.level] = lpv


def ladder_update_pv(ladder: dict[float, float], pvs: list[PV]) -> None:
    for pv in pvs:
        if not pv.volume:
            ladder.pop(pv.price, None)
        else:
            ladder[pv.price] = pv.volume


def ladder_update_mo(ladder: dict[float, float], mos: list[MatchedOrder]) -> None:
    for mo in mos:
        if not mo.size:
            ladder.pop(mo.price, None)
        else:
            ladder[mo.price] = mo.size


class RunnerOrderBook:
    __slots__ = (
        "available_to_back",
        "available_to_lay",
        "best_available_to_back",
        "best_available_to_lay",
        "best_display_available_to_back",
        "best_display_available_to_lay",
        "starting_price_back",
        "starting_price_lay",
        "starting_price_near",
        "starting_price_far",
        "traded",
        "last_traded_price",
        "total_volume",
        "handicap",
    )

    def __init__(self) -> None:
        self.available_to_back: dict[float, float] = {}
        self.available_to_lay: dict[float, float] = {}
        self.best_available_to_back: dict[int, LPV] = {}
        self.best_available_to_lay: dict[int, LPV] = {}
        self.best_display_available_to_back: dict[int, LPV] = {}
        self.best_display_available_to_lay: dict[int, LPV] = {}
        self.starting_price_back: dict[float, float] = {}
        self.starting_price_lay: dict[float, float] = {}
        self.starting_price_near: float | None = None
        self.starting_price_far: float | None = None
        self.traded: dict[float, float] = {}
        self.last_traded_price: float | None = None
        self.total_volume: float | None = None
        self.handicap: float | None = None

    def update(self, rc: RunnerChange) -> None:
        if rc.atb:
            ladder_update_pv(self.available_to_back, rc.atb)
        if rc.atl:
            ladder_update_pv(self.available_to_lay, rc.atl)
        if rc.batb:
            ladder_update_lpv(self.best_available_to_back, rc.batb)
        if rc.batl:
            ladder_update_lpv(self.best_available_to_lay, rc.batl)
        if rc.bdatb:
            ladder_update_lpv(self.best_display_available_to_back, rc.bdatb)
        if rc.bdatl:
            ladder_update_lpv(self.best_display_available_to_lay, rc.bdatl)
        if rc.spb:
            ladder_update_pv(self.starting_price_back, rc.spb)
        if rc.spl:
            ladder_update_pv(self.starting_price_lay, rc.spl)
        if rc.spn:
            self.starting_price_near = rc.spn
        if rc.spf:
            self.starting_price_far = rc.spf
        if rc.trd:
            ladder_update_pv(self.traded, rc.trd)
        if rc.ltp:
            self.last_traded_price = rc.ltp
        if rc.tv:
            self.total_volume = rc.tv
        if rc.hc:
            self.handicap = rc.hc

    def __repr__(self) -> str:
        data = {}
        for key in self.__slots__:
            val = getattr(self, key, None)
            if not val:
                continue
            if isinstance(val, dict):
                if key in ("available_to_back", "available_to_lay"):
                    val = "[" + ", ".join(f"{price:.4g}:{volume:.2f}" for price, volume in sorted(val.items())) + "]"
                else:
                    val = "[" + ", ".join(f"{lpv.price:.4g}:{lpv.volume:.2f}" for _, lpv in sorted(val.items())) + "]"
            elif isinstance(val, float):
                val = f"{val:.2f}"
            else:
                val = str(val)
            data[key] = val
        fields = ", ".join(f"{key}={val}" for key, val in sorted(data.items()))
        return f"<{type(self).__name__}{' ' if fields else ''}{fields}>"


class ChangeCache:
    clk: str | None = None
    initial_clk: str | None = None
    publish_time: int | None = None
    stream_unreliable: bool | None = False
    conflate_ms: int | None = None

    def update_meta(self, msg: MCM | OCM) -> None:
        if msg.initial_clk:
            self.initial_clk = msg.initial_clk
        if msg.clk:
            self.clk = msg.clk
        if msg.conflate_ms:
            self.conflate_ms = msg.conflate_ms
        self.publish_time = msg.pt
        self.stream_unreliable = msg.stream_unreliable
        if msg.ct == ChangeType.SUB_IMAGE and not msg.segment_type or msg.segment_type == SegmentType.SEG_START:
            self.clear()

    def clear(self) -> None:
        raise NotImplementedError()


class _DefaultDict(defaultdict):
    """
    The caches for single markets are supposed to be referenceable from outside the
    cache via weakrefs. So if the cache gets updated, it doesn't leave data pending.
    As defaultdict itself can't be referenced with a weakref, let's subclass it.
    """

    def __init__(self, **kwargs):
        super().__init__(type(self).default_factory, **kwargs)

    default_factory = dict


class MarketOrderBook(_DefaultDict):
    """Order book for a single market, collecting a bunch of RunnerOrderBooks."""

    default_factory = RunnerOrderBook  # type: ignore[assignment]


class MarketSubscriptionCache(ChangeCache):
    """
    Orderbook for all markets of a market change stream.

    Market subscriptions are always in the underlying exchange currency - GBP. The default roll-up for GBP
    is £1 for batb / batl and bdatb / bdatl, This means that stakes of less than £1 (or currency
    equivalent) are rolled up to the next available price on the odds ladder. For atb / atl there is
    no roll-up. Available volume is displayed at all prices including those with less than £2 available.
    """

    def __init__(self):
        self.order_book: defaultdict[str, MarketOrderBook] = defaultdict(MarketOrderBook)
        self.definitions: dict[str, MarketDefinition] = {}

    def clear(self) -> None:
        self.order_book.clear()
        self.definitions.clear()

    def update(self, mcm: MCM) -> None:
        self.update_meta(mcm)
        if mcm.is_heartbeat:
            return
        if not mcm.mc:
            return

        for mc in mcm.mc:
            if mc.img:
                self.order_book.pop(mc.id, None)
            if mc.market_definition:
                self.definitions[mc.id] = mc.market_definition
            if not mc.rc:
                continue
            for rc in mc.rc:
                self.order_book[mc.id][rc.id].update(rc)


class RunnerOrders:
    """
    New subscriptions: Will receive an initial image with only E - Executable orders (unmatched).
    Live subscriptions: Will receive a transient of the order to EC - Execution Complete as the
    order transits into that state (allowing you to remove the order from your cache).

    Please note: EXECUTION_COMPLETE (fully matched) orders are only returned when transitioning from
    EXECUTABLE to EXECUTION_COMPLETE. The full details of EXECUTION_COMPLETE orders can only be viewed
    using listCurrentOrders/listMarketBook using orderProjections.
    """

    __slots__ = (
        "matched_backs",
        "matched_lays",
        "unmatched_orders",
        "executed_orders",
        "handicap",
    )

    def __init__(self) -> None:
        self.matched_backs: dict[float, float] = {}
        self.matched_lays: dict[float, float] = {}
        self.unmatched_orders: dict[int, Order] = {}
        self.executed_orders: dict[int, Order] = {}  # Does only contain recently completed orders
        self.handicap: float | None = None

    def update(self, orc: OrderRunnerChange) -> None:
        if orc.hc:
            self.handicap = orc.hc
        if orc.uo:
            for uo in orc.uo:
                if uo.execution_complete:
                    self.unmatched_orders.pop(uo.id, None)
                    self.executed_orders[uo.id] = uo
                else:
                    self.unmatched_orders[uo.id] = uo
        if orc.mb:
            ladder_update_mo(self.matched_backs, orc.mb)
        if orc.ml:
            ladder_update_mo(self.matched_lays, orc.ml)


class MarketOrders(_DefaultDict):
    """All orders for a single market, collecting a bunch of RunnerOrders."""

    default_factory = RunnerOrders  # type: ignore[assignment]


class OrderSubscriptionCache(ChangeCache):
    """
    Order subscriptions are provided in the currency of the account that the orders are placed in.
    """

    def __init__(self):
        self.orders: defaultdict[str, MarketOrders] = defaultdict(MarketOrders)

    def clear(self) -> None:
        self.orders.clear()

    def update(self, ocm: OCM) -> None:
        self.update_meta(ocm)
        if ocm.is_heartbeat:
            return
        if not ocm.oc:
            return

        for oc in ocm.oc:
            if oc.full_image:
                self.orders.pop(oc.id, None)
            if oc.closed:
                # TODO: Call some hooks?
                self.orders.pop(oc.id, None)
            if not oc.orc:
                # on closed markets
                continue
            for orc in oc.orc:
                if orc.full_image:
                    self.orders[oc.id].pop(orc.id, None)
                self.orders[oc.id][orc.id].update(orc)



from enum import Enum

from betfair_parser.strenums import DocumentedEnum, LowerStrEnum, StrEnum, auto, doc


class EndpointType(LowerStrEnum):
    IDENTITY = auto()
    IDENTITY_CERT = auto()
    NAVIGATION = auto()
    HEARTBEAT = auto()
    ACCOUNTS = auto()
    BETTING = auto()
    SCORES = auto()
    NONE = auto()  # To avoid issues with mypy


class RegulatorCode(DocumentedEnum):
    MR_ESP = doc("SPANISH GAMBLING AUTHORITY")
    MR_INT = doc("GIBRALTER REGULATOR")
    MR_ITA = doc("AMMINISTRAZIONE AUTONOMA DEI MONOPOLI DI STATO")
    MR_NJ = doc("NJRC - NEW JERSEY RACING COMMISSION")
    MR_TGC = doc("THE TASMANIAN GAMING COMMISSION")


class OrderType(DocumentedEnum):
    LIMIT = doc("A normal exchange limit order for immediate execution")
    LIMIT_ON_CLOSE = doc("Limit order for the auction (SP)")
    MARKET_ON_CLOSE = doc("Market order for the auction (SP)")
    MARKET_AT_THE_CLOSE = doc("Undocumented, but found out there in the wild, likely a mistake by Betfair")


class OrderSide(StrEnum):
    BACK = auto()
    LAY = auto()


class OrderResponse(StrEnum):
    SUCCESS = auto()
    FAILURE = auto()


class OrderStatus(DocumentedEnum):
    PENDING = doc(
        "An asynchronous order is yet to be processed. Once the bet has been processed by the exchange "
        "(including waiting for any in-play delay), the result will be reported and available on the "
        "Exchange Stream API and API NG. Not a valid search criteria on MarketFilter."
    )
    EXECUTION_COMPLETE = doc("An order that does not have any remaining unmatched portion.")
    EXECUTABLE = doc("An order that has a remaining unmatched portion.")
    EXPIRED = doc(
        "The order is no longer available for execution due to its time in force constraint. "
        "In the case of FILL_OR_KILL orders, this means the order has been killed because it "
        "could not be filled to your specifications. Not a valid search criteria on MarketFilter."
    )


class JSONExceptionCode(DocumentedEnum):
    INVALID_JSON = doc(
        value=-32700,
        docstring=(
            "Invalid JSON was received by the server. An error occurred on the server while parsing the JSON text."
        ),
    )
    METHOD_NOT_FOUND = doc(value=-32601, docstring="Method not found")
    INVALID_PARAMETERS = doc(
        value=-32602, docstring="Problem parsing the parameters, or a mandatory parameter was not found"
    )
    JSON_RPC_ERROR = doc(value=-32603, docstring="Internal JSON-RPC error")


class APINGExceptionCode(DocumentedEnum):
    TOO_MUCH_DATA = doc("The operation requested too much data")
    INVALID_INPUT_DATA = doc("Invalid input data")
    INVALID_SESSION_INFORMATION = doc("The session token passed is invalid or expired")
    NO_APP_KEY = doc("An application key is required for this operation")
    NO_SESSION = doc("A session token is required for this operation")
    UNEXPECTED_ERROR = doc("An unexpected internal error occurred that prevented successful request processing.")
    INVALID_APP_KEY = doc("The application key passed is invalid")
    TOO_MANY_REQUESTS = doc("There are too many pending requests")
    SERVICE_BUSY = doc("The service is currently too busy to service this request")
    TIMEOUT_ERROR = doc("Internal call to downstream service timed out")
    REQUEST_SIZE_EXCEEDS_LIMIT = doc(
        "The request exceeds the request size limit. Requests are limited to a total of 250"
        "betId's/marketId's (or a combination of both)"
    )
    ACCESS_DENIED = doc(
        "The calling client is not permitted to perform the specific action e.g. the using a Delayed "
        "App Key when placing bets or attempting to place a bet from a restricted jurisdiction."
    )
    SERVICE_UNAVAILABLE = doc("Service is currently unavailable")  # Used only in RaceStatus


# This needs to  be here and not in accounts, as we need it for the error struct definition
class AccountAPINGExceptionCode(DocumentedEnum):
    INVALID_INPUT_DATA = doc("Invalid input data. Please check the format of your request.")
    INVALID_SESSION_INFORMATION = doc(
        "The session token hasn't been provided, is invalid or has expired. You must login "
        "again to create a new session token."
    )
    UNEXPECTED_ERROR = doc("An unexpected internal error occurred that prevented successful request processing.")
    INVALID_APP_KEY = doc("The application key passed is invalid or is not present.")
    SERVICE_BUSY = doc("The service is currently too busy to service this request.")
    TIMEOUT_ERROR = doc("The internal call to downstream service timed out.")
    DUPLICATE_APP_NAME = doc("Duplicate application name.")
    APP_KEY_CREATION_FAILED = doc(
        "Creating application key version has failed. Please check that your application name is unique "
        "and doesn't contain your Betfair username."
    )
    APP_CREATION_FAILED = doc("Application creation has been failed.")
    NO_SESSION = doc(
        "A session token header ('X-Authentication') has not been provided in the request. Please note: "
        "The same error is returned by the Keep Alive operation if the X-Authentication header is provided "
        "but the session value is invalid or if the session has expired."
    )
    NO_APP_KEY = doc("An application key header ('X-Application') has not been provided in the request.")
    SUBSCRIPTION_EXPIRED = doc("An application key is required for this operation.")
    INVALID_SUBSCRIPTION_TOKEN = doc("The subscription token provided doesn't exist.")
    TOO_MANY_REQUESTS = doc("Too many requests. For more details relating to this error please see FAQ's.")
    INVALID_CLIENT_REF = doc("Invalid length for the client reference.")
    WALLET_TRANSFER_ERROR = doc("There was a problem transferring funds between your wallets.")
    INVALID_VENDOR_CLIENT_ID = doc("The vendor client ID is not subscribed to this Application Key.")
    USER_NOT_SUBSCRIBED = doc(
        "The user making the request is not subscribed to the Application Key they are trying to perform "
        "the action on (e.g. creating an Authorisation Code)."
    )
    INVALID_SECRET = doc("The vendor making the request has provided a vendor secret that does not match our records.")
    INVALID_AUTH_CODE = doc("The vendor making the request has not provided a valid authorization code.")
    INVALID_GRANT_TYPE = doc(
        "The vendor making the request has not provided a valid grant_type, or the grant_type they have "
        "passed does not match the parameters (authCode/refreshToken)."
    )
    CUSTOMER_ACCOUNT_CLOSED = doc("A token could not be created because the customer's account is CLOSED.")


class EventTypeIdCode(int, Enum):
    """Unique identifier for event types

    This is not a complete list and only serves for ease of use. For a full list, use the ListEventTypes API call.

    https://docs.developer.betfair.com/display/1smk3cen4v3lu3yomq5qye0ni/Additional+Information?preview=/43090178/43090179/EventTypIds.xlsx
    """

    SOCCER = 1
    TENNIS = 2
    GOLF = 3
    CRICKET = 4
    RUGBY_UNION = 5
    BOXING = 6
    HORSE_RACING = 7
    MOTOR_SPORT = 8
    SPECIAL_BETS = 10
    CYCLING = 11
    ROWING = 12
    RUGBY_LEAGUE = 1477
    DARTS = 3503
    ATHLETICS = 3988
    GREYHOUND_RACING = 4339
    FINANCIAL_BETS = 6231
    SNOOKER = 6422
    AMERICAN_FOOTBALL = 6423
    BASEBALL = 7511
    BASKETBALL = 7522
    HOCKEY = 7523
    ICE_HOCKEY = 7524
    SUMO_WRESTLING = 7525
    AUSTRALIAN_RULES = 61420
    GAELIC_FOOTBALL = 66598
    HURLING = 66599
    POOL = 72382
    CHESS = 136332
    POKER_ROOM = 189929
    TROTTING = 256284
    COMMONWEALTH_GAMES = 300000
    POKER = 315220
    WINTER_SPORTS = 451485
    HANDBALL = 468328
    NETBALL = 606611
    SWIMMING = 620576
    BADMINTON = 627555
    INTERNATIONAL_RULES = 678378
    BRIDGE = 982477
    YACHTING = 998916
    VOLLEYBALL = 998917
    BOWLS = 998918
    BANDY = 998919
    FLOORBALL = 998920
    EXCHANGE_POKER = 1444073
    EXCHANGE_BLACKJACK = 1444076
    EXCHANGE_BACCARAT = 1444085
    EXCHANGE_HI_LO = 1444092
    EXCHANGE_OMAHA_HI = 1444099
    EXCHANGE_CARD_RACING = 1444115
    CASINO = 1444120
    EXCHANGE_ROULETTE = 1444130
    EXCHANGE_BULLSEYE_ROULETTE = 1444150
    BACKGAMMON = 1938544
    GAA_SPORTS = 2030972
    GAELIC_GAMES = 2152880
    INTERNAL_MARKETS = 2264869
    POLITICS = 2378961
    TABLE_TENNIS = 2593174
    YAHOO_RACING = 2791893
    BEACH_VOLLEYBALL = 2872194
    CANOEING = 2872196
    WATER_POLO = 2901849
    POLO = 2977000
    FISHING = 3088925
    ROLLER_HOCKEY = 3130721
    CROSS_SPORT_ACCUMULATORS = 3145419
    SQUASH = 4609466
    SURFING = 4726642
    COMBAT_SPORTS = 4968929
    EXCHANGE_GAMES = 5402258
    PELOTA = 5412697
    FEATURED_MARKETS = 5545197
    EXCHANGE_CASINO = 10271443
    TEN_PIN_BOWLING = 10390264
    TRADEFAIR = 15242720
    FUTSAL = 15826206
    FUSSBALL = 15826207
    HARNESS_RACING = 16872235
    EQUESTRIAN = 18643353
    HORSE_RACING_VIRTUAL = 26397698
    MIXED_MARTIAL_ARTS = 26420387
    TRIATHLON = 27065662
    CURRENT_AFFAIRS = 27388198
    VIRTUAL_SPORTS = 27438978
    E_SPORTS = 27454571
    WRESTLING = 27485048
    NEW_CUSTOMER_OFFER = 27596832
    UFC = 27610222
    GAA_FOOTBALL = 27610230
    GAA_HURLING = 27610231
    MMA_UFC = 27829360
    KABADDI = 27979456
    GYMNASTICS = 28347302
    TV_SPECIALS = 28347303
    MUSIC = 28347304
    NOVELTY_BETS = 28347305
    LOTTERY_SPECIALS = 28361978
    BEACH_SOCCER = 28361979
    HOLLYWOOD = 28361980
    WEIGHTLIFTING = 28361982
    WEATHER = 28361983
    SKI_JUMPING = 28361984
    NORDIC_COMBINED = 28361985
    ALPINE_SKIING = 28361987
    CROSS_COUNTRY = 28361988
    BIATHLON = 28361989
    FREESTYLE_SKIING = 28361990
    SPEED_SKATING = 28361992
    SPORTS_NOVELTIES = 28373540



from betfair_parser.spec.accounts.enums import (
    AffiliateRelationStatus,
    ItemClass,
    SubscriptionStatus,
    TokenType,
    Wallet,
    WinLose,
)
from betfair_parser.spec.common import BaseMessage, Date


class ApplicationSubscription(BaseMessage, frozen=True):
    """Application subscription details"""

    subscription_token: str  # Application key identifier
    expiry_date_time: Date | None = None
    expired_date_time: Date | None = None
    created_date_time: Date | None = None
    activation_date_time: Date | None = None
    cancellation_date_time: Date | None = None
    subscription_status: SubscriptionStatus | None = None
    client_reference: str | None = None
    vendor_client_id: str | None = None


class SubscriptionHistory(BaseMessage, frozen=True):
    """Application subscription history details"""

    subscription_token: str  # Application key identifier
    expiry_date_time: Date | None = None
    expired_date_time: Date | None = None
    created_date_time: Date | None = None
    activation_date_time: Date | None = None
    cancellation_date_time: Date | None = None
    subscription_status: SubscriptionStatus | None = None
    client_reference: str | None = None


class SubscriptionTokenInfo(BaseMessage, frozen=True):
    """Subscription token information"""

    subscription_token: str
    activated_date_time: Date | None = None
    expiry_date_time: Date | None = None
    expired_date_time: Date | None = None
    cancellation_date_time: Date | None = None
    subscription_status: SubscriptionStatus | None = None


class AccountSubscription(BaseMessage, frozen=True):
    """Application subscription details"""

    subscription_tokens: list[SubscriptionTokenInfo]
    application_name: str | None = None
    application_version_id: str | None = None


class DeveloperAppVersion(BaseMessage, frozen=True):
    """Describes a version of an external application"""

    owner: str  # The user who owns the specific version of the application
    version_id: int  # The unique Id of the application version
    version: str  # identifier string such as 1.0, 2.0. Unique for a given application.
    application_key: str  # The unique application key associated with this application version

    # Indicates whether the data exposed by platform services as seen by this
    # application key is delayed or realtime.
    delayData: bool
    subscription_required: bool  # Indicates whether the application version needs explicit subscription

    # Indicates whether the application version needs explicit management by the software owner.
    # A value of false indicates, this is a version meant for personal developer use.
    owner_managed: bool
    active: bool  # Indicates whether the application version is currently active

    # Public unique string provided to the Vendor that they can use to pass to the
    # Betfair API in order to identify themselves.
    vendor_id: str | None = None
    # Private unique string provided to the Vendor that they pass with certain calls
    # to confirm their identity. Linked to a particular App Key.
    vendor_secret: str | None = None


class DeveloperApp(BaseMessage, frozen=True):
    """Describes developer/vendor specific application"""

    app_name: str  # The unique name of the application
    app_id: int  # A unique id of this application
    app_versions: list[DeveloperAppVersion]  # The application versions (including application keys)


class AccountFundsResponse(BaseMessage, frozen=True):
    """Response for retrieving available to bet."""

    available_to_bet_balance: float | None = None
    exposure: float | None = None
    retained_commission: float | None = None  # Sum of retained commission.
    exposure_limit: float | None = None

    # User Discount Rate. Please note: Betfair AUS/NZ customers should not rely on this to determine
    # their discount rates which are now applied at the account level.
    discount_rate: float | None = None
    points_balance: int | None = None

    # This is not documented in the API description, but seems to be present anyway
    wallet: Wallet | None = None


class AccountDetailsResponse(BaseMessage, frozen=True):
    """Response for Account details."""

    # Default user currency Code. See Currency Parameters for minimum bet sizes relating to each currency.
    currency_code: str | None = None
    first_name: str | None = None
    last_name: str | None = None
    locale_code: str | None = None

    # Region based on users zip/postcode (ISO 3166-1 alpha-3 format). Defaults to GBR if
    # zip/postcode cannot be identified.
    region: str | None = None
    timezone: str | None = None  # User Time Zone.

    # User Discount Rate.   Please note:  Betfair AUS/NZ customers should not rely on this to
    # determine their discount rates which are now applied at the account level.
    discount_rate: float | None = None
    points_balance: int | None = None  # The Betfair points balance.
    country_code: str | None = None  # The customer's country of residence (ISO 2 Char format)


class StatementLegacyData(BaseMessage, frozen=True):
    """Summary of a cleared order."""

    avg_price: float | None  # The average matched price of the bet (null if no part has been matched)

    # The amount of the stake of your bet. (0 for commission payments or deposit/withdrawals)
    bet_size: float | None = None
    bet_type: str | None = None  # Back or lay
    bet_category_type: str | None = None  # Exchange, Market on Close SP bet, or Limit on Close SP bet.
    commission_rate: str | None = None  # Commission rate on market
    event_id: int | None = None  # Please note: this is the Id of the market without the associated exchangeId
    event_type_id: int | None = None

    # Full Market Name. For card payment items, this field contains the card name
    full_market_name: str | None = None
    gross_bet_amount: float | None = None  # The winning amount to which commission is applied.

    # Market Name. For card transactions, this field indicates the type of card
    # transaction (deposit, deposit fee, or withdrawal).
    market_name: str | None = None

    # Market type. For account deposits and withdrawals, marketType is set to NOT_APPLICABLE.
    market_type: str | None = None
    placed_date: Date | None = None  # Date and time of bet placement

    # Id of the selection (this will be the same for the same selection across markets)
    selection_id: int | None = None
    selection_name: str | None = None  # Name of the selection
    start_date: Date | None = None  # Date and time at the bet portion was settled
    transaction_type: str | None = None  # Debit or credit
    transaction_id: int | None = None  # The unique reference Id assigned to account deposit and withdrawals.
    win_lose: WinLose | None = None

    # In the instance of a dead heat, this field will indicate the number of winners
    # involved in the dead heat (null otherwise)
    dead_heat_price_divisor: float | None = None

    # Currently returns same value as avgPrice. Once released will display the average matched
    # price of the bet with no rounding applied
    avg_price_raw: float | None = None


class StatementItem(BaseMessage, kw_only=True, frozen=True):
    """Summary of a cleared order."""

    # An external reference, eg. equivalent to betId in the case of an exchange bet statement item.
    ref_id: str | None = None

    # The date and time of the statement item, eg. equivalent to settledData for an exchange
    # bet statement item. (in ISO-8601 format, not translated)
    item_date: Date
    amount: float | None = None  # The amount of money the balance is adjusted by
    balance: float | None = None

    # Class of statement item. This value will determine which set of keys will be included in itemClassData
    item_class: ItemClass | None = None
    # Key value pairs describing the current statement item. The set of keys will be
    # determined by the itemClass
    item_class_data: dict[str, str] | None = None
    # Set of fields originally returned from APIv6. Provided to facilitate migration from
    # APIv6 to API-NG, and ultimately onto itemClass and itemClassData
    legacy_data: StatementLegacyData | None = None


class AccountStatementReport(BaseMessage, frozen=True):
    """A container representing search results."""

    account_statement: list[StatementItem]  # The list of statement items returned by your request.
    more_available: bool  # Indicates whether there are further result items beyond this page.


class CurrencyRate(BaseMessage, frozen=True):
    currency_code: str | None = None  # Three-letter ISO 4217 code
    rate: float | None = None  # Exchange rate for the currency specified in the request


class AuthorisationResponse(BaseMessage, frozen=True):
    """Wrapper object containing authorisation code and redirect URL for web vendors"""

    authorisation_code: str  # The authorisation code
    redirect_url: str  # URL to redirect the user to the vendor page


class SubscriptionOptions(BaseMessage, frozen=True, rename=None):
    # No rename: SubscriptionOption fields don't use camelCase in Betfair API

    """Wrapper object containing details of how a subscription should be created"""

    # How many days should a created subscription last for. Open-ended subscription created
    # if value not provided. Relevant only if createdSubscription is true.
    subscription_length: int | None = None

    # An existing subscription token that the caller wishes to be activated instead of creating a new one.
    # Ignored is createSubscription is true.
    subscription_token: str | None = None
    client_reference: str | None = None  # Any client reference for this subscription token request.


class VendorAccessTokenInfo(BaseMessage, frozen=True, rename=None):
    # No rename: VendorAccessTokenInfo fields don't use camelCase in Betfair API

    """
    Wrapper object containing UserVendorSessionToken, RefreshToken and
    optionally a Subscription Token if one was created
    """

    access_token: str  # Session token used by web vendors
    token_type: TokenType  # Type of the token
    expires_in: int  # How long until the token expires
    refresh_token: str  # Token used to refresh the session token in future

    # Object containing the vendor client id and optionally some subscription information
    application_subscription: ApplicationSubscription


class VendorDetails(BaseMessage, frozen=True):
    """Wrapper object containing vendor name and redirect url"""

    app_version_id: int  # Internal id of the application
    vendor_name: str
    redirect_url: str | None = None  # URL to be redirected to


class AffiliateRelation(BaseMessage, frozen=True):
    """Wrapper object containing affiliate relation details"""

    status: AffiliateRelationStatus
    vendor_client_id: str  # ID of user



from functools import partial

from betfair_parser.spec.common import BaseMessage, Date, EndpointType, Params, Request, Response, method_tag
from betfair_parser.strenums import DocumentedEnum, doc


scores_tag = partial(method_tag, "ScoresAPING/v1.0/")


class RaceStatus(DocumentedEnum):
    DORMANT = doc("There is no data available for this race.")
    DELAYED = doc("The start of the race has been delayed")
    PARADING = doc("The horses/greyhounds are in the parade ring")
    GOINGDOWN = doc("The horses are going down to the starting post")
    GOINGBEHIND = doc("The horses are going behind the stalls")
    APPROACHING = doc("The greyhounds are approaching the traps")
    GOINGINTRAPS = doc("The greyhounds are being put in the traps")
    HARERUNNING = doc("The hare has been started")
    ATTHEPOST = doc("The horses are at the post")
    OFF = doc("The greyhound/horse race has started")
    FINISHED = doc("The race has finished")
    FINALRESULT = doc("The result has been declared (Greyhounds only)")
    FALSESTART = doc("There has been a false start")
    PHOTOGRAPH = doc("The result of the race is subject to a photo finish")
    RESULT = doc("The result of the race has been announced")
    WEIGHEDIN = doc("The jockeys have weighed in")
    RACEVOID = doc("The race has been declared void")
    NORACE = doc("The race has been declared a no race")
    MEETINGABANDONED = doc("The meeting has been abandoned")
    RERUN = doc("The race will be rerun")
    ABANDONED = doc("The race has been abandoned")


class ResponseCode(DocumentedEnum):
    OK = doc("Data returned successfully")
    NO_NEW_UPDATES = doc("No updates since the passes UpdateSequence")
    NO_LIVE_DATA_AVAILABLE = doc("Event scores are no longer available or are not on the schedule")
    SERVICE_UNAVAILABLE = doc("Data feed for the event type (tennis/football etc) is currently unavailable")
    UNEXPECTED_ERROR = doc("An unexpected error occurred retrieving score data")
    LIVE_DATA_TEMPORARILY_UNAVAILABLE = doc(
        "Live Data feed for this event/match is temporarily unavailable, data could potentially be stale"
    )


class RaceDetails(BaseMessage, kw_only=True, frozen=True):
    # Even as these items are marked as mandatory, they seem to be missing a lot
    meeting_id: str | None = None  # The unique Id for the meeting as returned by listEvents
    race_id: str | None = None  # The unique Id for the race in the format meetingId.raceTime (hhmm)
    race_status: RaceStatus | None = None  # The current status of the race.
    last_updated: Date | None = None  # This is the time the data was last updated
    sequence: int | None = None  # This is the unique identifier associated to each update of the data
    response_code: ResponseCode


class _ListRaceDetailsParams(Params, frozen=True):
    meeting_ids: set[str] | None = None  # Restricts the results to the specified meeting IDs.
    race_ids: set[str] | None = None  # Restricts the results to the specified race IDs.


class ListRaceDetails(Request, kw_only=True, frozen=True, tag=scores_tag):
    endpoint_type = EndpointType.SCORES
    params: _ListRaceDetailsParams
    return_type = Response[list[RaceDetails]]



import msgspec

from betfair_parser.spec.betting.enums import BetStatus, GroupBy, OrderBy, OrderProjection, Side, SortDir
from betfair_parser.spec.betting.type_definitions import (
    CancelExecutionReport,
    CancelInstruction,
    ClearedOrderSummaryReport,
    CurrentOrderSummaryReport,
    MarketVersion,
    PlaceExecutionReport,
    PlaceInstruction,
    ReplaceExecutionReport,
    ReplaceInstruction,
    RunnerId,
    UpdateExecutionReport,
    UpdateInstruction,
    betting_tag,
)
from betfair_parser.spec.common import (
    BetId,
    CustomerOrderRef,
    CustomerRef,
    CustomerStrategyRef,
    EndpointType,
    EventId,
    EventTypeId,
    MarketId,
    Params,
    Request,
    Response,
    TimeRange,
)


class _OrderRequest(Request, frozen=True, tag=betting_tag):
    endpoint_type = EndpointType.BETTING


class _PlaceOrdersParams(Params, frozen=True):
    market_id: str
    instructions: list[PlaceInstruction]
    customer_ref: CustomerRef | None = None
    market_version: MarketVersion | None = None
    customer_strategy_ref: CustomerStrategyRef | None = None
    async_: bool | None = msgspec.field(name="async", default=False)


class PlaceOrders(_OrderRequest, kw_only=True, frozen=True):
    """Place new orders into market.

    Please note that additional bet sizing rules apply to bets placed into the Italian Exchange.

    In normal circumstances the placeOrders is an atomic operation. PLEASE NOTE: if the
    'Best Execution' features is switched off, placeOrders can return ‘PROCESSED_WITH_ERRORS’
    meaning that some bets can be rejected and other placed when submitted in the same
    PlaceInstruction.
    """

    params: _PlaceOrdersParams
    return_type = Response[PlaceExecutionReport]


class _CancelOrdersParams(Params, frozen=True):
    market_id: str | None = None
    instructions: list[CancelInstruction] | None = None
    customer_ref: CustomerRef | None = None


class CancelOrders(_OrderRequest, kw_only=True, frozen=True):
    """
    Cancel all bets OR cancel all bets on a market OR fully or partially cancel particular
    orders on a market. Only LIMIT orders can be cancelled or partially cancelled once placed.
    """

    params: _CancelOrdersParams
    return_type = Response[CancelExecutionReport]


class _ReplaceOrdersParams(Params, frozen=True):
    market_id: str
    instructions: list[ReplaceInstruction]
    customer_ref: CustomerRef | None = None
    market_version: MarketVersion | None = None
    async_: bool | None = msgspec.field(name="async", default=False)


class ReplaceOrders(_OrderRequest, kw_only=True, frozen=True):
    """
    This operation is logically a bulk cancel followed by a bulk place. The cancel is completed
    first then the new orders are placed. The new orders will be placed atomically in that they
    will all be placed or none will be placed. In the case where the new orders cannot be placed
    the cancellations will not be rolled back. See ReplaceInstruction.
    """

    params: _ReplaceOrdersParams
    return_type = Response[ReplaceExecutionReport]


class _ListClearedOrdersParams(Params, frozen=True):
    bet_status: BetStatus  # Restricts the results to the specified status.
    event_type_ids: set[EventTypeId] | None = None  # Restricts the results to the specified Event Type IDs.
    event_ids: set[EventId] | None = None  # Restricts the results to the specified Event IDs.
    market_ids: set[MarketId] | None = None  # Restricts the results to the specified market IDs.
    runner_ids: set[RunnerId] | None = None  # Restricts the results to the specified Runners.
    bet_ids: set[BetId] | None = None  # Restricts the results to the specified bet IDs, maximum 1000 betIds
    customer_order_refs: set[CustomerOrderRef] | None = None
    customer_strategy_refs: set[CustomerStrategyRef] | None = None
    side: Side | None = None  # Restricts the results to the specified side.

    # Optionally restricts the results to be from/to the specified settled date. This date is inclusive,
    # i.e. if an order was cleared on exactly this date (to the millisecond) then it will be included
    # in the results. If 'from' is later than 'to', no results will be returned.
    # Please Note: if you have a longer running market that is settled at multiple different times
    # then there is no way to get the returned market rollup to only include bets settled in a certain
    # date range, it will always return the overall position from the market including all settlements.
    settled_date_range: TimeRange | None = None

    # If not supplied then the lowest level is returned, i.e. bet by bet This is only applicable to SETTLED BetStatus.
    group_by: GroupBy | None = None
    include_item_description: bool | None = None
    locale: str | None = None  # The language used for the itemDescription, defaults to account settings
    from_record: int | None = None  # Specifies the first record that will be returned. Records start at index zero.
    record_count: int | None = None  # Number of records from the index position 'fromRecord', maximum 1000


class ListClearedOrders(_OrderRequest, kw_only=True, frozen=True):
    """
    Returns a list of settled bets based on the bet status, ordered by settled date. To retrieve
    more than 1000 records, you need to make use of the fromRecord and recordCount parameters.

    By default the service will return all available data for the last 90 days (see Best Practice
    note below).  The fields available at each roll-up are available here

    Best Practice:
    You should specify a settledDateRange "from" date when making requests for data. This reduces
    the amount of data that requires downloading & improves the speed of the response. Specifying
    a "from" date of the last call will ensure that only new data is returned.
    """

    params: _ListClearedOrdersParams
    return_type = Response[ClearedOrderSummaryReport]


class _ListCurrentOrdersParams(Params, frozen=True):
    """
    Parameters for retrieving a list of current orders.
    """

    bet_ids: set[BetId] | None = None  # Restricts the results to the specified bet IDs
    market_ids: set[str] | None = None  # Restricts the results to the specified market IDs
    order_projection: OrderProjection | None = None  # Restricts the results to the specified order status
    customer_order_refs: set[CustomerOrderRef] | None = None
    customer_strategy_refs: set[CustomerStrategyRef] | None = None
    date_range: TimeRange | None = None  # Restricts the results to be from/to the specified date
    order_by: OrderBy | None = None  # Specifies how the results will be ordered
    sort_dir: SortDir | None = None  # Specifies the direction the results will be sorted in
    from_record: int | None = None  # Specifies the first record that will be returned
    record_count: int | None = None  # Specifies how many records will be returned
    include_item_description: bool | None = None


class ListCurrentOrders(_OrderRequest, kw_only=True, frozen=True):
    """
    Returns a list of your current orders. Optionally you can filter and sort your current orders
    using the various parameters, setting none of the parameters will return all of your current
    orders up to a maximum of 1000 bets, ordered BY_BET and sorted EARLIEST_TO_LATEST. To retrieve
    more than 1000 orders, you need to make use of the fromRecord and recordCount parameters.

    Best Practice:
    To efficiently track new bet matches from a specific time, customers should use a combination of
    the dateRange, orderBy "BY_MATCH_TIME" and orderProjection “ALL” to filter fully/partially matched
    orders from the list of returned bets. The response will then filter out any bet records that
    have no matched date and provide a list of betIds in the order which they are fully/partially
    matched from the date and time specified in the dateRange field.
    """

    params: _ListCurrentOrdersParams
    return_type = Response[CurrentOrderSummaryReport]


class _UpdateOrdersParams(Params, frozen=True):
    market_id: str  # The market id these orders are to be placed on
    instructions: list[UpdateInstruction]  # The limit of update instructions per request is 60
    customer_ref: CustomerRef | None = None


class UpdateOrders(_OrderRequest, kw_only=True, frozen=True):
    """Update non-exposure changing fields."""

    params: _UpdateOrdersParams
    return_type = Response[UpdateExecutionReport]



from typing import Annotated, Literal

import msgspec

from betfair_parser.spec.betting.enums import MarketBettingType, MarketStatus, MarketTypeCode, RunnerStatus
from betfair_parser.spec.common import (
    BaseMessage,
    BetId,
    CompetitionId,
    Date,
    EventId,
    EventTypeIdCode,
    Handicap,
    MarketId,
    Price,
    RegulatorCode,
    SelectionId,
    Size,
    Venue,
)
from betfair_parser.spec.streaming.enums import LapseStatusReasonCode, MarketDataFilterFields, PriceLadderDefinitionType


StreamRef = int | str

# Request objects


class MarketFilter(BaseMessage, frozen=True):
    betting_types: list[MarketBettingType] | None = None  # Match the betting type of the market
    bsp_market: bool | None = None  # If set, restrict to BSP or non-BSP markets only. If unset, return both
    country_codes: list[str] | None = None  # Restrict to specified country or countries. Defaults to 'GB' on error
    event_ids: list[EventId] | None = None  # Restrict markets by the event id associated with the market
    event_type_ids: list[EventTypeIdCode] | None = None  # Restrict markets by event type associated with the market
    market_ids: list[MarketId] | None = None  # If no marketIds passed user will be subscribed to all markets
    market_types: list[MarketTypeCode] | None = None  # Restrict to markets that match the type of the market
    race_types: list[str] | None = None  # Harness, Flat, Hurdle, Chase, Bumper, NH Flat, Steeple or NO_VALUE
    turn_in_play_enabled: bool | None = None  # If set, restrict to turn-inplay or non-inplay markets. Both if unset
    venues: list[Venue] | None = None  # Restrict by the venue associated with the market. Only for horse racing


class MarketDataFilter(BaseMessage, frozen=True):
    fields: list[MarketDataFilterFields] | None = None
    ladder_levels: int | None = None


class OrderFilter(BaseMessage, frozen=True):
    include_overall_position: bool = True  # Return overall position (See: OrderRunnerChange.mb / OrderRunnerChange.ml)
    customer_strategy_refs: list[str] | None = None  # Restricts to specified customerStrategyRefs

    # Returns strategy positions (See: OrderRunnerChange.smc=Map<customerStrategyRef, StrategyMatchChange>)
    # these are sent in delta format as per overall position
    partition_matched_by_strategy_ref: bool = False

    # Internal use only & should not be set on your filter (your subscription is already locked to your account).
    # If set subscription will fail.
    account_ids: list[int] | None = None


# Response objects


class RunnerDefinition(BaseMessage, frozen=True):
    sort_priority: int
    id: SelectionId
    name: str | None = None  # Undefined, but partly present
    hc: Handicap | None = None
    status: RunnerStatus | None = None
    adjustment_factor: float | None = None
    bsp: float | None = None
    removal_date: Date | None = None

    @property
    def handicap(self):
        """The handicap of the runner (selection) (None if not applicable)"""
        return self.hc


class KeyLineSelection(BaseMessage, frozen=True):
    id: int
    hc: float


class KeyLineDefinition(BaseMessage, frozen=True):
    kl: list[KeyLineSelection]


class PriceLadderDefinition(BaseMessage, frozen=True):
    type: PriceLadderDefinitionType


class MarketDefinition(BaseMessage, kw_only=True, frozen=True):
    bet_delay: int
    betting_type: MarketBettingType
    bsp_market: bool
    bsp_reconciled: bool
    competition_id: CompetitionId | None = None
    competition_name: str | None = None
    complete: bool
    country_code: str | None = None
    cross_matching: bool
    discount_allowed: bool | None = None
    each_way_divisor: float | None = None
    event_id: EventId
    event_name: str | None = None
    event_type_id: EventTypeIdCode
    in_play: bool
    key_line_definition: KeyLineDefinition | None = None

    # For Handicap and Line markets, the lines available on this market will be between the range of
    # lineMinUnit and lineMaxUnit, in increments of the lineInterval value. e.g. If unit is runs,
    # lineMinUnit=10, lineMaxUnit=20 and lineInterval=0.5, then valid lines include 10, 10.5, 11, 11.5 up to 20 runs.
    line_interval: float | None = None
    # For Handicap and Line markets, the maximum value for the outcome, in market units for this market (eg 100 runs).
    line_max_unit: float | None = None
    # For Handicap and Line markets, the minimum value for the outcome, in market units for this market (eg 0 runs).
    line_min_unit: float | None = None

    market_base_rate: float | None
    market_id: MarketId | None = None
    market_name: str | None = None
    market_time: Date
    market_type: str
    name: str | None = None
    number_of_active_runners: int
    number_of_winners: int
    open_date: Date | None = None
    persistence_enabled: bool
    price_ladder_definition: PriceLadderDefinition | PriceLadderDefinitionType | None = None
    race_type: str | None = None
    regulators: list[RegulatorCode]
    runners: list[RunnerDefinition]
    runners_voidable: bool
    settled_time: Date | None = None
    status: MarketStatus
    suspend_time: Date
    timezone: str | None = None
    turn_in_play_enabled: bool
    venue: str | None = None
    version: int | None = None

    @property
    def event_type_name(self) -> str:
        return self.event_type_id.name


class PV(BaseMessage, array_like=True, frozen=True):
    """Price-Volume pair"""

    price: Price
    volume: Size


class LPV(BaseMessage, array_like=True, frozen=True):
    """Level-Price-Volume triple"""

    level: int
    price: Price
    volume: Size


AvailableToBack = Annotated[PV, msgspec.Meta(title="AvailableToBack")]
AvailableToLay = Annotated[PV, msgspec.Meta(title="AvailableToLay")]
BestAvailableToBack = Annotated[LPV, msgspec.Meta(title="BestAvailableToBack")]
BestAvailableToLay = Annotated[LPV, msgspec.Meta(title="BestAvailableToLay")]
BestDisplayAvailableToBack = Annotated[LPV, msgspec.Meta(title="BestDisplayAvailableToBack")]
BestDisplayAvailableToLay = Annotated[LPV, msgspec.Meta(title="BestDisplayAvailableToLay")]
StartingPriceBack = Annotated[PV, msgspec.Meta(title="StartingPriceBack")]
StartingPriceLay = Annotated[PV, msgspec.Meta(title="StartingPriceLay")]
Trade = Annotated[PV, msgspec.Meta(title="Trade")]


class RunnerChange(BaseMessage, frozen=True):
    id: SelectionId
    atb: list[AvailableToBack] | None = None
    atl: list[AvailableToLay] | None = None
    batb: list[BestAvailableToBack] | None = None
    batl: list[BestAvailableToLay] | None = None
    bdatb: list[BestDisplayAvailableToBack] | None = None
    bdatl: list[BestDisplayAvailableToLay] | None = None
    spb: list[StartingPriceBack] | None = None  # Starting Price (Available To) Back
    spl: list[StartingPriceLay] | None = None  # Starting Price (Available To) Lay
    spn: float | None = None  # Starting Price Near
    spf: float | None = None  # Starting Price Far
    trd: list[Trade] | None = None  # Traded
    ltp: float | None = None  # Last Traded Price
    tv: float | None = None  # Total Volume
    hc: Handicap | None = None

    @property
    def available_to_back(self):
        """PriceVol tuple delta of price changes (0 vol is remove)"""
        return self.atb

    @property
    def available_to_lay(self):
        """PriceVol tuple delta of price changes (0 vol is remove)"""
        return self.atl

    @property
    def best_available_to_back(self):
        """LevelPriceVol triple delta of price changes, keyed by level (0 vol is remove)"""
        return self.batb

    @property
    def best_available_to_lay(self):
        """LevelPriceVol triple delta of price changes, keyed by level (0 vol is remove)"""
        return self.batl

    @property
    def best_display_available_to_back(self):
        """LevelPriceVol triple delta of price changes, keyed by level (0 vol is remove) (includes virtual prices)"""
        return self.bdatb

    @property
    def best_display_available_to_lay(self):
        """LevelPriceVol triple delta of price changes, keyed by level (0 vol is remove) (includes virtual prices)"""
        return self.bdatl

    @property
    def starting_price_back(self):
        """PriceVol tuple delta of price changes (0 vol is remove)"""
        return self.spb

    @property
    def starting_price_lay(self):
        """PriceVol tuple delta of price changes (0 vol is remove)"""
        return self.spl

    @property
    def starting_price_near(self):
        """The near starting price (or None if un-changed)"""
        return self.spn

    @property
    def starting_price_far(self):
        """The far starting price (or None if un-changed)"""
        return self.spf

    @property
    def traded(self):
        """PriceVol tuple delta of price changes (0 vol is remove)"""
        return self.trd

    @property
    def last_traded_price(self):
        """The last traded price (or None if un-changed)"""
        return self.ltp

    @property
    def total_volume(self):
        """The total amount matched. This value is truncated at 2dp."""
        return self.tv

    @property
    def handicap(self):
        """The handicap of the runner (selection) (None if not applicable)"""
        return self.hc


class MarketChange(BaseMessage, kw_only=True, frozen=True):
    id: MarketId
    rc: list[RunnerChange] | None = None  # Runner Changes
    con: bool | None = None  # Conflated
    img: bool = False  # Image
    market_definition: MarketDefinition | None = None
    tv: float | None = None  # Traded Volume

    @property
    def runner_changes(self):
        """Not present if market_definition is sent"""
        return self.rc

    @property
    def conflated(self):
        """Have more than a single change been combined (or None if not conflated)"""
        return self.con

    @property
    def image(self):
        """Replace existing prices / data with the data supplied: it is not a delta"""
        return self.img

    @property
    def traded_volume(self):
        """Total amount matched across the market. None if un-changed"""
        return self.tv


class Order(BaseMessage, frozen=True):
    id: BetId
    p: float  # Price
    s: float  # Size
    # Side of the order. For Line markets a 'B' bet refers to a SELL line and an 'L' bet refers to a BUY line.
    side: Literal["B", "L"]
    status: Literal["E", "EC"]  # Status of the order (E = EXECUTABLE, EC = EXECUTION_COMPLETE)
    pt: Literal["L", "P", "MOC"]  # Persistence Type
    ot: Literal["L", "MOC", "LOC"]  # Order Type - codespell-ignore
    pd: int  # Placed Date
    bsp: float | None = None  # BSP Liability
    rfo: str | None = None  # Order Reference
    rfs: str | None = None  # Strategy Reference
    rc: str | None = None  # Regulator Code
    rac: str | None = None  # Regulator Auth Code
    # TODO: convert int(ms) into datetime for dates??
    md: int | None = None  # Matched Date
    cd: int | None = None  # Cancelled Date
    ld: int | None = None  # Lapsed Date
    avp: float | None = None  # Average Price Matched
    sm: float | None = None  # Size Matched
    sr: float | None = None  # Size Remaining
    sl: float | None = None  # Size Lapsed
    sc: float | None = None  # Size Cancelled
    sv: float | None = None  # Size Voided
    lsrc: LapseStatusReasonCode | None = None

    @property
    def execution_complete(self) -> bool:
        return self.status == "EC"

    @property
    def price(self):
        """
        The original placed price of the order. Line markets operate at even-money odds of 2.0.
        However, price for these markets refers to the line positions available as defined by the markets
        min-max range and interval steps
        """
        return self.p

    @property
    def size(self):
        """The original placed size of the order"""
        return self.s

    @property
    def persistence_type(self):
        """If the order will persist at in play or not (L=LAPSE, P=PERSIST, MOC=Market On Close)"""
        return self.pt

    @property
    def order_type(self):
        """The type of the order (L = LIMIT, MOC = MARKET_ON_CLOSE, LOC = LIMIT_ON_CLOSE)"""
        return self.ot  # codespell-ignore

    @property
    def placed_date(self):
        """The date the order was placed (in millis since epoch) that the changes were generated"""
        return self.pd

    @property
    def bsp_liability(self):
        """The BSP liability of the order (None if the order is not a BSP order)"""
        return self.bsp

    @property
    def customer_order_ref(self):
        """The customer's order reference for this order (None if one was not set)"""
        return self.rfo

    @property
    def customer_strategy_ref(self):
        """The customer's strategy reference for this order (empty string if one was not set)"""
        return self.rfs

    @property
    def regulator_code(self):
        """The regulator of the order"""
        return self.rc

    @property
    def regulator_auth_code(self):
        """The auth code returned by the regulator"""
        return self.rac

    @property
    def matched_date(self):
        """The date the order was matched (None if the order is not matched)"""
        return self.md

    @property
    def cancelled_date(self):
        """The date the order was cancelled (None if the order is not cancelled)"""
        return self.cd

    @property
    def lapsed_date(self):
        """The date the order was lapsed (None if the order is not lapsed)"""
        return self.ld

    @property
    def average_price_matched(self):
        """
        The average price the order was matched at (None if the order is not matched). This value
        is not meaningful for activity on Line markets and is not guaranteed to be returned or
        maintained for these markets.
        """
        return self.avp

    @property
    def size_matched(self):
        """The amount of the order that has been matched"""
        return self.sm

    @property
    def size_remaining(self):
        """The amount of the order that is remaining unmatched"""
        return self.sr

    @property
    def size_lapsed(self):
        """The amount of the order that has been lapsed"""
        return self.sl

    @property
    def size_cancelled(self):
        """The amount of the order that has been cancelled"""
        return self.sc

    @property
    def size_voided(self):
        """The amount of the order that has been voided"""
        return self.sv

    @property
    def lapse_status_reason_code(self):
        """The reason that some or all of this order has been lapsed (None if no portion of the order is lapsed"""
        return self.lsrc

    @property
    def bet_id(self):
        # interchangeability with betting.Order
        return self.id

    avg_price_matched = average_price_matched  # interchangeability with betting.Order


class MatchedOrder(BaseMessage, array_like=True, frozen=True):
    price: Price
    size: Size


class StrategyMatchChange(BaseMessage, frozen=True):
    mb: list[MatchedOrder] | None = None  # Matched Backs
    ml: list[MatchedOrder] | None = None  # Matched Lays

    @property
    def matched_backs(self):
        """Matched amounts by distinct matched price on the Back side for this strategy"""
        return self.mb

    @property
    def matched_lays(self):
        """Matched amounts by distinct matched price on the Lay side for this strategy"""
        return self.ml


class OrderRunnerChange(BaseMessage, frozen=True):
    id: SelectionId
    full_image: bool | None = False
    hc: Handicap | None = None
    mb: list[MatchedOrder] | None = None  # Matched Backs
    ml: list[MatchedOrder] | None = None  # Matched Lays
    smc: dict[str, StrategyMatchChange] | None = None  # Strategy Matches
    uo: list[Order] | None = None  # Unmatched Orders

    @property
    def handicap(self):
        """The handicap of the runner (selection) (None if not applicable)"""
        return self.hc

    @property
    def matched_backs(self):
        """Matched amounts by distinct matched price on the Back side for this runner (selection)"""
        return self.mb

    @property
    def matched_lays(self):
        """Matched amounts by distinct matched price on the Lay side for this runner (selection)"""
        return self.ml

    @property
    def strategy_matches(self):
        """Matched Backs and Matched Lays grouped by strategy reference"""
        return self.smc

    @property
    def unmatched_orders(self):
        """Orders on this runner (selection) that are not fully matched"""
        return self.uo


class OrderMarketChange(BaseMessage, kw_only=True, frozen=True):
    id: MarketId
    account_id: int | None = None
    closed: bool | None = None
    full_image: bool | None = False
    orc: list[OrderRunnerChange] | None = None

    @property
    def order_runner_changes(self):
        """A list of changes to orders on a selection"""
        return self.orc



from betfair_parser.spec.accounts import enums, operations, type_definitions
from betfair_parser.spec.accounts.enums import (
    AffiliateRelationStatus,
    GrantType,
    IncludeItem,
    ItemClass,
    MarketType,
    Status,
    SubscriptionStatus,
    TokenType,
    Wallet,
    WinLose,
)
from betfair_parser.spec.accounts.operations import (
    GetAccountDetails,
    GetAccountFunds,
    GetAccountStatement,
    ListCurrencyRates,
)
from betfair_parser.spec.accounts.type_definitions import (
    AccountDetailsResponse,
    AccountFundsResponse,
    AccountStatementReport,
    AccountSubscription,
    AffiliateRelation,
    ApplicationSubscription,
    AuthorisationResponse,
    CurrencyRate,
    DeveloperApp,
    DeveloperAppVersion,
    StatementItem,
    StatementLegacyData,
    SubscriptionHistory,
    SubscriptionOptions,
    SubscriptionTokenInfo,
    VendorAccessTokenInfo,
    VendorDetails,
)
from betfair_parser.spec.common import TimeRange  # noqa



from itertools import count
from typing import Any, ClassVar, Generic, Literal, TypeVar, get_type_hints

import msgspec

from betfair_parser.exceptions import APIError, APINGException, JSONError
from betfair_parser.spec.common.enums import (
    AccountAPINGExceptionCode,
    APINGExceptionCode,
    EndpointType,
    JSONExceptionCode,
)


def decode(raw: bytes, type: Any = Any) -> Any:
    try:
        return msgspec.json.decode(raw, strict=False, type=type)
    except msgspec.DecodeError as e:
        raise JSONError(str(e)) from e


def encode(data: Any) -> bytes:
    try:
        return msgspec.json.encode(data)
    except msgspec.EncodeError as e:
        raise JSONError(str(e)) from e


def first_lower(s: str) -> str:
    """Lower only the first character of a string."""
    return s[:1].lower() + s[1:]


def method_tag(prefix: str, class_name: str) -> str:
    return prefix + first_lower(class_name)


class BaseMessage(
    msgspec.Struct,
    kw_only=True,
    forbid_unknown_fields=True,
    omit_defaults=True,
    repr_omit_defaults=True,
    frozen=True,
    rename="camel",
):
    @classmethod
    def parse(cls, raw):
        return decode(raw, type=cls)

    def validate(self):
        return bool(decode(encode(self), type=type(self)))

    def to_dict(self):
        return msgspec.structs.asdict(self)

    def replace(self, **kwargs):
        return msgspec.structs.replace(self, **kwargs)


class Params(BaseMessage, frozen=True):
    """
    Base class for request parameters. Don't send None and other redundant default
    values. If not subclassed, this class is used to describe an empty parameter set.
    """


class BaseResponse(BaseMessage, frozen=True):
    """Base class for Response, ErrorResponse and identity responses."""

    @property
    def is_error(self):
        """True, if the response is some kind of error."""
        return False


class RPC(BaseMessage, omit_defaults=False, repr_omit_defaults=False, frozen=True):
    jsonrpc: Literal["2.0"] = "2.0"
    id: int = 1


ResultType = TypeVar("ResultType")
ParamsType = TypeVar("ParamsType")
ErrorCode = TypeVar("ErrorCode")


class ExceptionDetails(BaseMessage, Generic[ErrorCode], kw_only=True, frozen=True):
    error_code: ErrorCode
    error_details: str | None = None  # The stack trace of the error
    request_uuid: str | None = msgspec.field(name="requestUUID", default=None)

    @property
    def code(self) -> ErrorCode:
        return self.error_code


class ExceptionData(BaseMessage, frozen=True, rename=None):
    exception_name: Literal["APINGException", "AccountAPINGException"] = msgspec.field(name="exceptionname")
    APINGException: ExceptionDetails[APINGExceptionCode] | None = None
    AccountAPINGException: ExceptionDetails[AccountAPINGExceptionCode] | None = None

    @property
    def error(self) -> ExceptionDetails:
        return getattr(self, self.exception_name)

    @property
    def code(self) -> str:
        return self.error.code


class RPCError(BaseMessage, frozen=True):
    code: int  # JSONExceptionCode or any other undocumented integer code, if it's an APINGException
    message: str  # Something like "DSC-0018", "AANGX-0011", ...
    data: ExceptionData | None = None  # The interesting part

    @property
    def exception_code(self):
        """Return some form of ExceptionCode, whatever error we get."""
        return self.data.code if self.data else JSONExceptionCode(self.code)

    def __str__(self) -> str:
        return str(self.exception_code)


class Response(RPC, BaseResponse, Generic[ResultType], kw_only=True, frozen=True):
    """RPC response. Either an error or a result."""

    result: ResultType | None = None
    error: RPCError | None = None

    @property
    def is_error(self) -> bool:
        return self.error is not None


_default_id_generator = count(1)


def _failsafe_issubclass(x: Any, A: type) -> bool:
    try:
        return issubclass(x, A)
    except TypeError:
        return False


class Request(RPC, kw_only=True, frozen=True, tag_field="method", tag=first_lower):
    # class variables for subclassing, which msgspec won't serialize in messages
    endpoint_type: ClassVar[EndpointType] = EndpointType.NONE
    return_type: ClassVar[type] = Response
    throws: ClassVar[type] = APINGException  # JSON error definition

    # Having no default for `params` makes sure, that initializing a message object that
    # requires parameters without explicitly calling the `with_params` constructor fails.
    params: Params

    @classmethod
    def _params_cls(cls) -> type:
        """Get the according parameter class from the type hints."""
        hinted_type = get_type_hints(cls)["params"]
        if _failsafe_issubclass(hinted_type, Params):
            # params: Params
            return hinted_type
        if hinted_type is type(None):
            # params: None
            return dict
        if hinted_type.__args__:
            # params: Optional[Params]
            hinted_type = hinted_type.__args__[0]
            if _failsafe_issubclass(hinted_type, Params):
                return hinted_type
        return dict

    @classmethod
    def with_params(cls, request_id=None, **kwargs):
        """General constructor for RPC requests."""
        params = cls._params_cls()(**kwargs)
        if request_id is None:
            request_id = next(_default_id_generator)
        return cls(
            params=params,
            id=request_id,
        )

    @property
    def method(self) -> str:
        """Return the RPC request method as defined by the classes tag configuration."""
        return str(self.__struct_config__.tag)

    @staticmethod
    def headers() -> dict[str, str]:
        """HTTP headers of the request."""
        return {
            "Accept": "application/json",
            "Content-Type": "application/json",
            # X-Application to be set by the application
            # X-Authentication to be set by the application
        }

    def body(self) -> bytes:
        """HTTP body of the request."""
        return encode(self)

    def parse_response(self, response: bytes, raise_errors: bool = True):
        """Parse the received response into the defined return type."""
        resp = decode(response, type=self.return_type)
        if resp.id != self.id:
            raise APIError(f"Response ID ({resp.id}) does not match Request ID ({self.id})")
        if not resp.is_error:
            return resp.result
        if not raise_errors:
            return resp.error
        raise self.throws(
            str(resp.error.exception_code),
            response=resp,
            request=self,
            code=resp.error.exception_code,
        )



import asyncio
import io
import itertools
import pathlib
import socket
import ssl
import urllib.parse
from collections.abc import AsyncGenerator, Callable, Iterable
from typing import Any

from betfair_parser.cache import MarketSubscriptionCache, OrderSubscriptionCache
from betfair_parser.exceptions import StreamError
from betfair_parser.spec.common import encode
from betfair_parser.spec.streaming import (
    MCM,
    OCM,
    Authentication,
    ChangeMessageType,
    Connection,
    Heartbeat,
    MarketSubscription,
    OrderSubscription,
    Status,
    StreamRef,
    StreamResponseType,
    SubscriptionType,
    stream_decode,
)


LINE_SEPARATOR = b"\r\n"


def _default_handler(msg: StreamResponseType) -> StreamResponseType:
    return msg


class ExchangeStream:
    """Handle the byte stream with betfair."""

    def __init__(self, app_key: str, token: str, id_generator: Callable | None = None) -> None:
        self.app_key = app_key
        self.token = token
        self.subscriptions: dict[StreamRef, SubscriptionType] = {}
        self.handlers: dict[StreamRef, Callable] = {}
        self._id_generator = id_generator if id_generator is not None else itertools.count(1000)
        self._connection_id: str | None = None

    @property
    def connection_id(self) -> str | None:
        return self._connection_id

    @property
    def is_connected(self) -> bool:
        return bool(self.connection_id)

    def unique_id(self) -> int:
        return next(self._id_generator)  # type: ignore[call-overload]

    def handle_connection(self, msg: Connection) -> Connection:
        self._connection_id = msg.connection_id
        return msg

    def handle_heartbeat(self, msg: Heartbeat) -> Heartbeat:  # noqa
        return msg

    def handle_status(self, msg: Status) -> Status:
        if msg.is_error or msg.connection_closed:
            raise StreamError(
                f"Connection {self.connection_id} to stream {msg.id} failed: {msg.error_code}: {msg.error_message}"
            )
        return msg

    def handle_msg(self, msg: StreamResponseType) -> Any:
        match msg:
            case Heartbeat():
                return self.handle_heartbeat(msg)
            case Status():
                return self.handle_status(msg)
            case Connection():
                return self.handle_connection(msg)
        try:
            return self.handlers[msg.id](msg)
        except KeyError:
            raise StreamError(f"Unexpected stream message: {msg}")
        except Exception as e:
            raise StreamError(f"Handling stream message failed: {msg}") from e

    def authenticate(self) -> bytes:
        return encode(Authentication(id=self.unique_id(), app_key=self.app_key, session=self.token)) + LINE_SEPARATOR

    def subscribe(self, subscription: SubscriptionType, handler: Callable = _default_handler) -> bytes | None:
        self.subscriptions[subscription.id] = subscription
        self.handlers[subscription.id] = handler
        if self.connection_id:
            # only write something, if the connection is already established
            return encode(subscription) + LINE_SEPARATOR
        return None

    def connect(self) -> bytes:
        auth = self.authenticate()
        if not self.subscriptions:
            return auth

        # send out subscriptions, that were registered before connecting
        subscriptions = LINE_SEPARATOR.join(encode(subscription) for subscription in self.subscriptions.values())
        return auth + subscriptions + LINE_SEPARATOR

    def receive_bytes(self, data: bytes) -> Any:
        if not data:
            return None
        return self.handle_msg(stream_decode(data))  # type: ignore[arg-type]

    def receive(self, stream: io.RawIOBase) -> Any:
        return self.receive_bytes(stream.readline())


def create_ssl_socket(hostname, timeout: float | None = None) -> ssl.SSLSocket:
    """Create ssl socket and set timeout."""
    context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    context.load_default_certs()
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    secure_sock = context.wrap_socket(s, server_hostname=hostname)
    secure_sock.settimeout(timeout)
    return secure_sock


def create_stream_io(endpoint, timeout: float = 15):
    """Open an IO stream through a TLS connection to the given endpoint."""
    url = urllib.parse.urlparse(endpoint)
    sock = create_ssl_socket(url.hostname, timeout=timeout)
    sock.connect((url.hostname, url.port))
    return socket.SocketIO(sock, "rwb")


def changed_markets(msg: StreamResponseType) -> list[str]:
    """Return the market IDs of the markets affected by the given message."""
    if isinstance(msg, MCM) and msg.market_changes:
        return [m.id for m in msg.market_changes]
    if isinstance(msg, OCM) and msg.order_market_changes:
        return [m.id for m in msg.order_market_changes]
    return []


class StreamReader:
    """Read exchange stream data into a separate cache for each subscription."""

    def __init__(self, app_key, token) -> None:
        self.caches: dict[StreamRef, MarketSubscriptionCache | OrderSubscriptionCache] = {}
        self.esm = ExchangeStream(app_key, token)

    def handle_change_message(self, msg: ChangeMessageType) -> ChangeMessageType:
        self.caches[msg.id].update(msg)  # type: ignore[arg-type]
        return msg

    def subscribe(self, subscription: SubscriptionType) -> bytes:
        if isinstance(subscription, MarketSubscription):
            self.caches[subscription.id] = MarketSubscriptionCache()
        elif isinstance(subscription, OrderSubscription):
            self.caches[subscription.id] = OrderSubscriptionCache()
        else:
            raise TypeError("Invalid subscription type")
        return self.esm.subscribe(subscription, self.handle_change_message)

    def receive(self, stream: io.RawIOBase) -> Any:
        return self.esm.receive(stream)

    def connect(self, stream: io.RawIOBase) -> None:
        self.esm.receive(stream)  # read connection
        stream.write(self.esm.connect())  # send auth
        self.esm.receive(stream)

    def iter_changes(self, stream: io.RawIOBase) -> Iterable[ChangeMessageType]:
        """Iterate over the stream, yielding market and order change messages."""
        if not self.esm.is_connected:
            self.connect(stream)

        while True:
            msg = self.esm.receive(stream)
            if not msg:
                return
            if isinstance(msg, ChangeMessageType):  # type: ignore[arg-type,misc]
                yield msg

    def iter_changes_and_write(
        self,
        stream: io.RawIOBase,
        path: pathlib.Path | str,
    ) -> Iterable[ChangeMessageType]:
        if not self.esm.is_connected:
            self.connect(stream)

        with open(path, "ab") as f:
            while True:
                raw_msg = stream.readline()
                if not raw_msg:
                    return
                msg = self.esm.receive_bytes(raw_msg)
                if isinstance(msg, ChangeMessageType):  # type: ignore[arg-type,misc]
                    yield msg
                f.write(raw_msg)


class AsyncStream:
    """Async version of io.RawIOBase over a SSL connection."""

    _reader: asyncio.StreamReader | None = None
    _writer: asyncio.StreamWriter | None = None

    def __init__(self, endpoint, timeout: float = 15) -> None:
        self._endpoint = endpoint
        self._timeout = timeout

    async def connect(self) -> None:
        url = urllib.parse.urlparse(self._endpoint)
        context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        context.load_default_certs()
        self._reader, self._writer = await asyncio.open_connection(
            host=url.hostname,
            port=url.port,
            ssl=context,
            server_hostname=url.hostname,
            limit=1_000_000,
        )

    async def write(self, data: bytes) -> None:
        if not self._writer:
            raise StreamError("Stream is not connected")
        self._writer.write(data)
        await self._writer.drain()

    async def readline(self) -> bytes:
        if not self._reader:
            raise StreamError("Stream is not connected")
        return await self._reader.readline()

    async def close(self) -> None:
        self._reader = None  # does not need to be closed explicitly
        if self._writer:
            await self._writer.drain()

            # Abort the underlying transport (equivalent to socket.shutdown())
            transport = self._writer.transport
            if transport:
                transport.abort()

            # Close the writer (which should in turn close the underlying transport)
            self._writer.close()
            await self._writer.wait_closed()
        self._writer = None

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()


class AsyncStreamReader(StreamReader):
    async def receive_async(self, stream: AsyncStream) -> Any:
        return self.esm.receive_bytes(await stream.readline())

    async def connect_async(self, stream: AsyncStream) -> None:
        self.esm.receive_bytes(await stream.readline())  # read connection
        await stream.write(self.esm.connect())  # send auth
        self.esm.receive_bytes(await stream.readline())

    async def iter_changes_async(self, stream: AsyncStream) -> AsyncGenerator[ChangeMessageType, None]:
        if not self.esm.is_connected:
            await self.connect_async(stream)

        while True:
            msg = self.esm.receive_bytes(await stream.readline())
            if not msg:
                return
            if isinstance(msg, ChangeMessageType):  # type: ignore[arg-type,misc]
                yield msg

    async def iter_changes_and_write_async(
        self,
        stream: AsyncStream,
        path: pathlib.Path | str,
    ) -> AsyncGenerator[ChangeMessageType, None]:
        if not self.esm.is_connected:
            await self.connect_async(stream)

        loop = asyncio.get_running_loop()
        with open(path, "ab") as f:
            while True:
                raw_msg = await stream.readline()
                if not raw_msg:
                    return
                msg = self.esm.receive_bytes(raw_msg)
                if isinstance(msg, ChangeMessageType):  # type: ignore[arg-type,misc]
                    yield msg
                await loop.run_in_executor(None, f.write, raw_msg)



"""Shared definitions between the different API parts."""

from betfair_parser.spec.common.enums import (
    AccountAPINGExceptionCode,
    APINGExceptionCode,
    EndpointType,
    EventTypeIdCode,
    JSONExceptionCode,
    OrderResponse,
    OrderSide,
    OrderStatus,
    OrderType,
    RegulatorCode,
)
from betfair_parser.spec.common.messages import (
    BaseMessage,
    BaseResponse,
    Params,
    Request,
    Response,
    decode,
    encode,
    first_lower,
    method_tag,
)
from betfair_parser.spec.common.type_definitions import (
    BetId,
    CompetitionId,
    CountryCode,
    CustomerOrderRef,
    CustomerRef,
    CustomerStrategyRef,
    Date,
    EventId,
    EventTypeId,
    ExchangeId,
    Handicap,
    IDType,
    MarketId,
    MarketType,
    MatchId,
    Price,
    SelectionId,
    Size,
    TimeRange,
    Venue,
)



import re
from typing import Literal, Union

from betfair_parser.spec.common import BaseMessage, BaseResponse, Date, EndpointType, Request


def navigation_tag(class_name: str):
    """
    >>> navigation_tag("EventType")
    'EVENT_TYPE'

    >>> navigation_tag("Group")
    'GROUP'
    """
    return re.sub(r"(?<!^)(?=[A-Z])", "_", class_name).upper()


class Market(BaseMessage, tag=navigation_tag, frozen=True):
    name: str
    id: str
    exchange_id: str
    market_type: str
    market_start_time: Date
    number_of_winners: int | str


class Event(BaseMessage, tag=navigation_tag, frozen=True):
    name: str
    id: int
    country_code: str
    children: list[Union["Group", "Event", Market]]


class Race(BaseMessage, tag=navigation_tag, frozen=True):
    name: str
    id: str
    country_code: str
    venue: str
    start_time: Date
    children: list[Market]
    race_number: str | None = None


class Group(BaseMessage, tag=navigation_tag, frozen=True):
    name: str
    id: str
    children: list[Union["Group", Event]]


class EventType(BaseMessage, tag=navigation_tag, frozen=True):
    name: str
    id: int
    children: list[Group | Event | Race]


class Navigation(BaseResponse, frozen=True):
    """Navigation root"""

    type: Literal["GROUP"]
    name: Literal["ROOT"]
    id: int
    children: list[EventType]

    @property
    def result(self):
        # Play along with ordinary RPC Response objects
        return self


class Menu(Request, kw_only=True, frozen=True, tag=""):
    """
    Navigation request
    https://docs.developer.betfair.com/display/1smk3cen4v3lu3yomq5qye0ni/Navigation+Data+For+Applications
    """

    endpoint_type = EndpointType.NAVIGATION
    params: None = None
    id: int = 0
    return_type = Navigation

    @classmethod
    def with_params(cls, request_id=None, **kwargs):
        # Force id to 0 to avoid issues with message validation
        return cls(id=0, **kwargs)

    def body(self):
        return b""


class FlattenedMarket(BaseMessage, kw_only=True, frozen=True, rename=None):
    event_type_name: str
    event_type_id: int
    event_name: str | None = None
    event_id: str | None = None
    event_country_code: str | None = None
    market_name: str
    market_id: str
    market_exchange_id: str
    market_market_type: str
    market_market_start_time: Date
    market_number_of_winners: int | str
    group_name: str | None = None
    group_id: str | None = None
    race_name: str | None = None
    race_id: str | None = None
    race_country_code: str | None = None
    race_venue: str | None = None
    race_start_time: Date | None = None
    race_race_number: str | None = None


def flatten_nav_tree(navigation: Navigation, **filters) -> list[FlattenedMarket]:
    return list(filter_flattened(flattened_nav_iter(navigation), **filters))


def _filter(k, v):
    if isinstance(v, str):
        return k == v
    if isinstance(v, tuple | list):
        return k in v
    raise TypeError


def filter_flattened(nav_iter, **filters):
    for flattened in nav_iter:
        if all(_filter(getattr(flattened, k), v) for k, v in filters.items()):
            yield flattened


def flattened_nav_iter(node, **context):
    node_type_name = navigation_tag(node.__class__.__name__).lower()
    context[node_type_name] = node
    if isinstance(node, Market):
        flattened = _flattened_from_context(context)
        yield flattened
    else:
        for child in node.children:
            yield from flattened_nav_iter(child, **context)


def _flattened_from_context(ctx):
    ctx.pop("navigation")
    return FlattenedMarket(
        **{
            f"{node_type}_{k}": v
            for node_type, nav_item in ctx.items()
            for k, v in nav_item.to_dict().items()
            if k not in ("type", "children")
        }
    )
